# This is a package to do simple manipulation of an Apache .htaccess file
# It manages a set of 'DENY FROM' lines with IP addresses.
# On update, the controlled 'deny from' lines are rewritten with the set
# of IP addresses. These are sorted and uniqued and evenly distributed across
# the controlled lines. New lines are added if necessary to prevent lines
# from having more than the maximum addresses per line.

use strict;

package Transfinitum::Apache::IPAccess;
use Fcntl qw(:DEFAULT :flock);

use constant MAX_ADDRS_PER_LINE => 50; # maximum number of IP addresses per line in the file
our $MARKER_LINE = '# -- MT AUTO BAN';

# An RE that matches IP addresses
our $IP_RE = qr/(?:[[:digit:]]+\.){3}[[:digit:]]+/;

# Create a new handler for the file.
# Argument: the path to the file
sub new {
    my ($class, $path) = (shift, shift);
    my %options = @_;
    my (%addrs,@lines,@addr_lines,@marks,$file);
    if (sysopen $file,$path,O_RDWR|O_CREAT) {
        # If client wants a lock and we don't get it, return undef as failure.
        return undef if ($options{lock} and not flock($file, LOCK_EX|LOCK_NB));
        # Load the file contents and close out the handle
        seek $file, 0, 0;
        @lines = <$file>;
        # If we're not locking, close out the file now.
        if (not $options{lock}) {
            close $file;
            $file = undef;
        }
        # Find our section of the file and make a note of the start and end
        # lines.
        my $i = -1;
        @marks = map { ++$i; m/^$MARKER_LINE/o ? $i : () } @lines;
        if (scalar @marks == 0) {
            @marks = ( scalar @lines, scalar @lines );
        } elsif ( scalar @marks == 1) {
            push(@marks, scalar @lines)
        } else {
            # remove any marker lines past the first two. Be sure to do this
            # starting from the last line, so the indices don't get bollixed
            splice(@lines, pop(@marks), 1) while (scalar @marks > 2);
        }

        # process the lines between the marks
        if (not $options{no_load}) {
            for ( my $idx = $marks[0] + 1 ; $idx < $marks[1] ; ++$idx ) {
                my $line = $lines[$idx];
                next if $line !~ m/^[[:space:]]*deny from/i;
                $line =~ s/($IP_RE)/&{sub { my $a = $_[0]; $addrs{$a} = 1; $a; }}($1)/geo;
            }
        }
        # remove all lines relating to our data.
        splice(@lines, $marks[0], $marks[1] - $marks[0] + 1);
    }
    # Extracted all the data we want, create the object and stash the data there.
    my $self = { addrs => \%addrs, # The set of banned addrs
                lines => \@lines, # The file content
                base_idx => $marks[0], # where to insert our content
                path => $path, # the path to the file
                fh => $file, # preserve in case we have a lock.
                };
    return bless $self, $class;
}

# Add an address to the set of banned addresses. Returns true if the address
# wasn't already banned
sub ban {
    my ($self, $addr) = @_;
    my $addr_set = $self->{addrs};
    return 0 if $addr !~ m/^$IP_RE$/o; # client tried to give us an invalid address
    return 0 if $addr_set->{$addr}; # it's already banned
    $addr_set->{$addr} = -1; # add it, mark it as not being on any line
    $self->{is_modified} = 1; # file needs to be written
    return 1;
}

# Test if in address is already banned
sub is_banned {
    my ($self, $addr) = @_;
    return exists $self->{addrs}->{$addr};
}

# Return the number of addresses banned
sub count {
    my ($self) = @_;
    return scalar(keys(%{$self->{addrs}}));
}

# Rewrite the file (if necessary) to store all of the banned addresses
# Options:
#   tag -> text to put on the end of the leading marker line
#   url -> url to sand trap file, if any.
#   rewrite -> if true, then mod_rewrite is used for redirection.
#       otherwise, if url is set, then ErrorDocument is used.
#       otherwise, it's just denial.
sub update {
    my ($self) = shift;
    my %options = @_;
    if ($self->{is_modified}) {
        # Cache data from the object
        my $addrs = $self->{addrs};
        my $lines = $self->{lines};
        my $url = $options{url};
        my $per_line = $options{per_line} || MAX_ADDRS_PER_LINE;

        # Generate a sorted list of addresses because humans like to see the
        # addresses in the access file in sorted order. We play funny map
        # games to make this as fast as possible. Primarily we want to avoid
        # doing a split on every comparison, so we do the split up front
        # then sort that and then compress back down to the address.
        my @a = map { $_->[0] }
                sort { $a->[1] <=> $b->[1] or $a->[2] <=> $b->[2] or $a->[3] <=> $b->[3] or $a->[4] <=> $b->[4] }
                map { [ $_, map { int($_) } split /\./ ] } keys(%$addrs);

        # put the addrs in rows
        my @addr_lines = ( $MARKER_LINE . ' (' . scalar(@a) . ' addresses) ' . ($options{tag}||'') ); # first marker line, with tag appended
        push (@addr_lines, 'ErrorDocument 403 '. $url) if ($url and not $options{rewrite});

        # Layout the working lines
        if (@a) {
            if ($url and $options{rewrite}) {
                push(@addr_lines, 'RewriteEngine On');
                # Drop the addresses in to chunky lines. We splice $per_line
                # addresses off the front of the array every loop and slap
                # them all in to one output line until we've done all the
                # addresses.
                while (@a) {
                    push(@addr_lines, 'RewriteCond %{REMOTE_ADDR} ^'
                                    . join('|' => splice(@a,0,$per_line))
                                    . '$ [OR]');
                }
                # Need to clip the '[OR]' from the last line. We checked that
                # we had addresses earier so there's always at least one line.
                $addr_lines[$#addr_lines] =~ s/\[OR\]$//o;
                push(@addr_lines, 'RewriteRule ^(.*) ' . $url);
            } else {
                while (@a) {
                    push (@addr_lines, 'deny from ' . join(' ' => splice(@a,0,$per_line)));
                }
            }
        }

        push(@addr_lines, $MARKER_LINE);
        $_ .= "\n" foreach @addr_lines;
        splice(@$lines, $self->{base_idx}, 0, @addr_lines);

        # Time to write the file
        my $file;
        # Use the cached handle if present, otherwise open the file.
        if ($self->{fh}) {
            $file = delete $self->{fh}; # pull it out so we can close it
            seek $file, 0, 0; # need seek between read/write operations.
            truncate $file, 0; # clear out anything there now.
        } else {
            open $file,'>',$self->{path};
        }

        if ($file) {
            print $file @$lines;
            close $file;
        }
        delete $self->{is_modified};
    }
}

1;

__END__

# test code

package main;

my $htaccess = Transfinitum::Apache::IPAccess->new('ht.access');

$htaccess->ban('1.1.1.5');
$htaccess->update();

1;


